#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

FW_NFP_CONFIGAGENT_TOPIC = 'nfp-firewall-agent'
LB_NFP_CONFIGAGENT_TOPIC = 'nfp-lbaas-agent'
LBV2_NFP_CONFIGAGENT_TOPIC = 'nfp-lbaasv2-agent'
VPN_NFP_CONFIGAGENT_TOPIC = 'nfp-vpn_agent'
NFP_NSO_TOPIC = "nfp-service-orchestrator"

FW_NFP_PLUGIN_TOPIC = 'q-firewall-plugin'
LB_NFP_PLUGIN_TOPIC = 'n-lbaas-plugin'
LBV2_NFP_PLUGIN_TOPIC = 'n-lbaasv2-plugin'
VPN_NFP_PLUGIN_TOPIC = 'vpn_plugin'
DEVICE_ORCH_TOPIC = 'nfp-configurator-ndo'
SERVICE_ORCH_TOPIC = 'nfp-configurator-nso'
CONFIG_ORCH_TOPIC = 'nfp-nco-notification-topic'
